// import 'source-map-support/register';
import { Random, sleep, Session } from 'koishi';
import { ThesaurusPluginConfig } from './config';
import { promises as fs } from 'fs';
import _ from 'lodash';
import {
  StarterPlugin,
  DefinePlugin,
  Inject,
  LifecycleEvents,
  UseMiddleware,
  UseCommand,
  CommandUsage,
  CommandExample,
  PutOption,
  PutUserName,
  PutSession,
} from 'koishi-thirdeye';
import AragamiPlugin, { CacheKey } from 'koishi-plugin-cache-aragami';
import { Next } from 'koa';

class ChatSession {
  @CacheKey()
  sessionId: string;
  name: string;
  playedWords: string[];
}

export * from './config';

@DefinePlugin({ name: 'thesaurus', schema: ThesaurusPluginConfig })
export default class ThesaurusPlugin
  extends StarterPlugin(ThesaurusPluginConfig)
  implements LifecycleEvents
{
  @Inject(true)
  private aragami: AragamiPlugin;

  private wordData: Record<string, string[]> = {};
  private async loadWords(): Promise<void> {
    const path = this.config.path;
    try {
      if (path.match(/^https?:\/\//)) {
        this.wordData = await this.ctx.http.get(path);
      } else {
        this.wordData = JSON.parse(await fs.readFile(path, 'utf-8'));
      }
      this.ctx
        .logger('thesaurus')
        .info(
          `Loaded ${
            Object.keys(this.wordData).length
          } patterns from source ${path} .`,
        );
    } catch (e) {
      this.ctx
        .logger('thesaurus')
        .error(`Failed to load words from source ${path}: ${e.toString()}`);
      await sleep(5000);
      return this.loadWords();
    }
  }

  private async getFromCache(sessionId: string) {
    if (!sessionId) {
      return;
    }
    return this.aragami.get(ChatSession, sessionId);
  }
  private async setToCache(chatSession: ChatSession) {
    return this.aragami.set(ChatSession, chatSession);
  }
  private async clearCacheOf(sessionId: string) {
    return this.aragami.del(ChatSession, sessionId);
  }

  private async replyWith(word: string, chatSession: ChatSession) {
    if (!word) {
      return;
    }
    chatSession.playedWords.push(word);
    if (chatSession.playedWords.length > this.config.trackingLength) {
      chatSession.playedWords.shift();
    }
    await this.setToCache(chatSession);
    return word.replace(/你/g, chatSession.name).trim();
  }

  private tryPatterns(patterns: string[], usedWords: Set<string>) {
    if (!patterns.length) {
      return;
    }
    const pattern = Random.pick(patterns);
    const availableWords = this.wordData[pattern].filter(
      (w) => !usedWords.has(w),
    );
    if (!availableWords.length) {
      const shrinkedPatterns = patterns.filter((p) => p !== pattern);
      return this.tryPatterns(shrinkedPatterns, usedWords);
    }
    return Random.pick(availableWords);
  }

  private async replyChat(word: string, chatSession: ChatSession) {
    const matchingPatterns = Object.keys(this.wordData).filter((w) =>
      word.includes(w),
    );
    if (!matchingPatterns.length) {
      const allWords = _.flatten(Object.values(this.wordData));
      let remainingAllWords = allWords.filter(
        (w) => !chatSession.playedWords.includes(w),
      );
      if (!remainingAllWords.length) {
        remainingAllWords = allWords;
      }
      return this.replyWith(Random.pick(remainingAllWords), chatSession);
    }
    const replyWord =
      this.tryPatterns(
        matchingPatterns,
        new Set<string>(chatSession.playedWords),
      ) || this.tryPatterns(matchingPatterns, new Set<string>());
    return this.replyWith(replyWord, chatSession);
  }
  private formatName(name: string) {
    if (name.match(/^[a-zA-Z0-9]/)) {
      name = ` ${name}`;
    }
    if (name.match(/[a-zA-Z0-9]$/)) {
      name = `${name} `;
    }
    return name;
  }

  @UseMiddleware()
  async replyMessage(session: Session, next: Next) {
    const sessionId = `${session.platform}.${session.selfId}.${session.userId}`;
    const chatSession = await this.getFromCache(sessionId);
    if (chatSession) {
      if (session.content === 'quit') {
        await this.clearCacheOf(sessionId);
        return '记得下次再找我聊喔~';
      }
      return await this.replyChat(session.content, chatSession);
    }
    return next();
  }

  @UseCommand('startchat', '开始聊天')
  @CommandUsage('聊天开始后， quit 停止聊天。')
  @CommandExample(
    'startchat --name Shigma 可以使用 Shigma 作为自己的名字来聊天。',
  )
  async startChat(
    @PutOption('name', '--name <name:string>') optionName: string,
    @PutUserName() databaseName: string,
    @PutSession() session: Session,
  ) {
    const sessionId = `${session.platform}.${session.selfId}.${session.userId}`;
    const name = optionName || databaseName;
    const formattedName = this.formatName(name);
    await this.setToCache({
      sessionId,
      name: formattedName,
      playedWords: [],
    });
    return `${formattedName}，开始和我聊天吧。聊累了的话，输入 quit 就可以结束聊天哦。`.trim();
  }

  async onConnect() {
    await this.loadWords();
  }
}
